package chord

import (
	"errors"
	"net/netip"
	"sync"
)

// Node represents a node's topological relations within a Chord network.
// It is safe to use the methods of Node concurrently.
type Node struct {
	// mu guards the mutable fields of the node.
	mu sync.Mutex
	// self is the node's own address.
	// It is fully immutable; the mutex does not protect it.
	self Peer
	// pred is the node's predecessor.
	// If not valid, the predecessor is not currently known.
	pred Peer
	// succ is the node's immediate successors for replication. Its capacity
	// is the replication factor for the network.
	// There is always at least one successor. If the number of nodes in the
	// network is not larger than the replication factor, then this node will
	// appear in its own successor list.
	succ []Peer
	// fingers is the node's finger table.
	fingers []Peer

	// data is the data the node owns.
	data map[ID]string
}

func (n *Node) Self() Peer {
	return n.self
}

func (n *Node) Successor() Peer {
	n.mu.Lock()
	defer n.mu.Unlock()
	return n.succ[0]
}

// Neighbors returns the node's predecessor and appends its successor list to s.
func (n *Node) Neighbors(s []Peer) (Peer, []Peer) {
	n.mu.Lock()
	defer n.mu.Unlock()
	return n.pred, append(s, n.succ...)
}

func (n *Node) succOwnsLocked(id ID) bool {
	return contains(n.self.id, n.succ[0].id, id)
}

// Closest finds the locally known peer which is the closest predecessor of key.
// If the boolean value is true, then this node believes that the peer is the
// closest predecessor in the network.
func (n *Node) Closest(id ID) (Peer, bool) {
	self := n.self.id
	n.mu.Lock()
	defer n.mu.Unlock()
	// Check the immediate successor first to satisfy the contract.
	if n.succOwnsLocked(id) {
		return n.succ[0], true
	}
	l := n.fingers
	for i := len(l) - 1; i >= 0; i-- {
		f := l[i]
		if contains(self, id, f.id) {
			// contains reports true if id == f.id, which is
			// technically incorrect for this operation.
			// We could reasonably skip double-checking and rely on the hash,
			// but we can also not skip it and just be correct.
			if id == f.id {
				continue
			}
			return f, false
		}
	}
	// Also try successors.
	l = n.succ
	for i := len(l) - 1; i >= 0; i-- {
		f := l[i]
		if contains(self, id, f.id) {
			if id == f.id {
				continue
			}
			return f, false
		}
	}
	// No known node is closer than we are.
	return n.self, true
}

// GetLocal obtains the value for a key.
func (n *Node) GetLocal(k ID) (v string, found bool) {
	n.mu.Lock()
	defer n.mu.Unlock()
	v, found = n.data[k]
	return v, found
}

// SetLocal sets the value for a key.
func (n *Node) SetLocal(k ID, v string) {
	n.mu.Lock()
	defer n.mu.Unlock()
	n.data[k] = v
}

// Leave tells the node that a peer is leaving.
// succ gives the leaving peer's successors.
func (n *Node) Leave(p Peer, succ []Peer) {
	n.mu.Lock()
	defer n.mu.Unlock()
	// The leaving peer might be either this node's successor, or its
	// predecessor, or both in the case of a network of two nodes.
	if p.addr == n.pred.addr {
		n.pred = Peer{}
	}
	for i, s := range n.succ {
		// TODO(branden): we should explicitly merge the successor lists
		if s.addr == p.addr {
			// Found the successor.
			// Delete it and add in the last received successor.
			copy(n.succ[i:], n.succ[i+1:])
			n.succ[len(n.succ)-1] = succ[len(succ)-1]
			break
		}
	}
}

// SuccessorFailed marks the node's current successor as having failed, e.g.
// during stabilization.
func (n *Node) SuccessorFailed() {
	// TODO(branden): probably most of this will need rewritten once
	// replication and fingers are implemented
	n.mu.Lock()
	defer n.mu.Unlock()
	old := n.succ[0]
	// If there are only two nodes in the network, the predecessor is also the
	// successor, which means it too has failed. Clear it in that case.
	if old.addr == n.pred.addr {
		n.pred = Peer{}
	}
	if len(n.succ) > 1 {
		// The successor list has replication. Just shift it down.
		copy(n.succ, n.succ[1:])
		n.succ = n.succ[:len(n.succ)-1]
	} else {
		// Check the finger table.
		for _, f := range n.fingers {
			if f.addr != n.self.addr {
				// TODO(branden): ensure the list stays sorted
				n.succ[0] = f
				break
			}
		}
	}
	if n.succ[0] != old {
		return
	}
	// We couldn't find a new successor in either the successor list or the
	// finger table. The only other candidate we have is our predecessor.
	// Stabilization will eventually work us out even if that's wrong.
	if n.pred.IsValid() {
		n.succ[0] = n.pred
		return
	}
	n.succ[0] = n.self
}

// Peer is the ID and address of a node.
type Peer struct {
	id   ID
	addr netip.AddrPort
}

// Address constructs a peer from an address and port.
func Address(addr netip.AddrPort) Peer {
	return Peer{id: addrID(addr), addr: addr}
}

// IsValid reports whether the peer was created using [Address] with an address
// which was valid.
func (p Peer) IsValid() bool { return p.addr.IsValid() && p.addr.Port() != 0 }

// Values returns the peer's ID and address.
func (p Peer) Values() (ID, netip.AddrPort) {
	// We do this instead of exporting Peer's fields to ensure those fields
	// are never mutable. If p.IsValid() then p.id == addrID(p.addr) always.
	return p.id, p.addr
}

// String formats the peer's address for debugging.
func (p Peer) String() string {
	return p.addr.String()
}

// Create creates a new Chord network using the given address as the initial node.
func Create(addr netip.AddrPort) (*Node, error) {
	if !addr.IsValid() {
		return nil, errors.New("chord: cannot create with invalid address")
	}
	self := Address(addr)
	n := &Node{
		self:    self,
		succ:    []Peer{self}[:1:1], // extra cautious about capacity
		fingers: make([]Peer, 0, len(ID{})*8),
		data:    make(map[ID]string),
	}
	return n, nil
}
